// DRAWOBJ.CPP	 The basic object type on the game map
//
// DJM

#include "scimain.h"
#include "drawobj.h"
#include "mapplane.h"
#include "gamemap.h"
#include "objectdb.h"
#include "procs.h"
#include "colordef.h"
#include "unit.h"
#include "player.h"
#include "debug.h"
#include "mapedit.h"
#include "obstaclepoint.h"
#include "cycler.h"
#include "missile.h"

DrawObj::DrawObj(MapPlane *poPlane, int nTemplateID)
{
	assert(poPlane);
	meClass = OC_DRAW;

	mpoPlane = poPlane;
	mpoCurNode = NULL;
	mbSelected = FALSE;
	mbSelectable = TRUE;
	mnCel = 0;
	mnLoop = 0;
	mnLayer = 0;
	mpoTemplate = NULL;
	mpoBitmap = NULL;
	mpoCycler = NULL;
	mbFlip = FALSE;

	mbDoneCycling = TRUE;
	mnObjID = ID_NONE;

	mjAlpha = 255;

	mpwSpecialPalette = NULL;

	mpoTemplate = (Template*)pApp->pResourceManager->LoadResource(RES_TEMPLATE, nTemplateID);
	assert(mpoTemplate);

	vSetBitmap(mpoTemplate->lGetValue(TAG_Bnor, nTemplateID));
	mbVisible = mpoTemplate->lGetValue(TAG_bVis, TRUE);
	
	goObjectDB.vAddObject(this);

	int nCycType = mpoTemplate->lGetValue(TAG_Atyp, 0);
	switch (nCycType)
	{
		case 0:
			break;
		case 1:
			mpoCycler = new ForwardCycler(this);
			break;
		case 2:
			mpoCycler = new ReverseCycler(this);
			break;
		case 3:
			mpoCycler = new OscillateCycler(this);
			break;
		case 4:
			mpoCycler = new RandomCycler(this);
			break;
		case 5:
			mpoCycler = new MotionCycler(this);
			break;
		default:
			break;
	}

	mfMaxHealth = (float)mpoTemplate->fGetValue(TAG_mhlt, 1000.0);
	mfHealth = mfMaxHealth;
	mpoChildren = NULL;
	mpoParent = NULL;
	meDrawState = (DRAW_STATE)mpoTemplate->lGetValue(TAG_drst, DS_ALWAYS);	//(only relevant for buildings)
	mbEventBubbling = mpoTemplate->lGetValue(TAG_ebub,1);		//whether the child passes up events to its parent
	mbParentOpen = FALSE;
	vInitChildren();

}


DrawObj::~DrawObj()
{
	goObjectDB.vRemoveObject(mnObjID);

	if (mpoBitmap) {
		pApp->pResourceManager->FreeResource(mpoBitmap);
		mpoBitmap = NULL;
	}
	if (mpoTemplate) {
		pApp->pResourceManager->FreeResource(mpoTemplate);
		mpoTemplate = NULL;
	}

	mpoPlane->vRemoveFromDrawnList(this);
	if (mpoCurNode) {
		mpoCurNode->vRemoveObj(this);	//remove self from node's drawn list
	}

	if (mpoCycler) {
		delete mpoCycler;
		mpoCycler = NULL;
	}
	
	//dispose child/parent relationships
	if (mpoParent) {
		mpoParent->vRemoveChild(this);
		mpoParent = NULL;
	}
	if (mpoChildren) {
		while(mpoChildren->GetSize())
		{
			DrawObj *poObj = mpoChildren->GetItem(0);
			//mpoChildren->Remove(poObj);
			delete poObj;
		}
		delete mpoChildren;
		mpoChildren = NULL;
	}

}

long DrawObj::lGetChecksum(void)
{
	// checksum all member variables that are important to net synch

	long lCheck = 0;

	// do self
	lCheck += (long)meClass + 
		nTemplateID();

	// do not include 
	//		mnLayer
	//		mnObjID
	//		mbSelectable
	return lCheck;
}

void DrawObj::vDumpSyncText(FILE *fileDump)
{
	// do self
	fprintf(fileDump, "*DrawObj\n");
	fprintf(fileDump, " Selectable?: %d\n", (int)mbSelectable);
	fprintf(fileDump, " Class: %d\n", (int)meClass);
	fprintf(fileDump, " Layer: %d\n", mnLayer);
	fprintf(fileDump, " ObjID: %d\n", mnObjID);
	fprintf(fileDump, " Template: %d\n", nTemplateID());
}

void DrawObj::vPlayRandomSound(long lTag, BOOL bWithObject)
{
	long *plList;
	long lSize;

	plList = mpoTemplate->plGetList(lTag, &lSize);

	if (lSize && plList)
	{
		int nWhich = random((short)lSize);
		int nSound = (int)plList[nWhich];

		if (bWithObject)
		{
			pApp->pSoundManager->EmitFromObject(this, nSound, SOUNDPRIORITY_UNITSOUNDS, FALSE, 128, TRUE);
		}
		else
		{
			pApp->pSoundManager->EmitFromPosition(nPixelX(), nPixelY(), nSound);
		}
	}
}

void DrawObj::vSetBitmap(int nBitmap)
{
	if (nBitmap == nBitmapID()) {
		return;
	}
	if (mpoBitmap) {
		pApp->pResourceManager->FreeResource(mpoBitmap);
		mpoBitmap = NULL;
	}
	
	mpoBitmap = (Bitmap*)pApp->pResourceManager->LoadResource(RES_BITMAP, nBitmap);
	//mpoOutsideBitmap = mpoBitmap;

	vMarkUpdate();
}

void DrawObj::vSetNode(MapNode *poNode)
{
	//assert(poNode);
	if (poNode == mpoCurNode) {
		return;
	}
	
	if (mpoCurNode) {
		//changing nodes
		if (bChildren())							//if any children
		{
			vSetChildNodeOff(poNode);				//then set children's nodes based on poNode
		}
		mpoCurNode->vRemoveObj(this);
	} else {
		//first time inserting into node
		if (bChildren())				//if any children
		{
			vSetChildNodeNew(poNode);	//then set nodes based on 'ctno' offsets
		}
	}
	mpoCurNode = poNode;
}

void DrawObj::vMoveToNode(MapNode *poNode)
{
	//assert(poNode);

	if (poNode == mpoCurNode) {
		return;
	}
	if (mpoCurNode) {
		mpoCurNode->vRemoveObj(this);
	}
	mpoCurNode = poNode;
	if (mpoCurNode) {
		mpoCurNode->vAddObj(this);
	}
}

void DrawObj::vCommand(Command &oCommand, BOOL bInternal)
{
	//NULL
}

Boolean DrawObj::bOnMe(sciEvent &oEvent)
{
	if ((oEvent.x >= mrCurrent.left) &&
		(oEvent.x <= mrCurrent.right) &&
		(oEvent.y >= mrCurrent.top) &&
		(oEvent.y <= mrCurrent.bottom)) {

		POINT op;
		op.x = oEvent.x - mrCurrent.left;
		op.y = oEvent.y - mrCurrent.top;
		mpoBitmap->SetLoop(mnLoop);	//make sure to do this so the 'shared' bitmap is currently where we need it to be
//		vDebugOut ("Color:%d\n", mpoBitmap->ColorOfPoint(op));
		if (mpoBitmap->ColorOfPoint(op) != 0) {
			return TRUE;
		} else {
			return FALSE;
		}

	} else {
		return FALSE;
	}
}

void DrawObj::vUpdateObstacles() {
	//get map size
	int nWidth  = mpoPlane->mpoMap->nGetWidth();
	int nHeight = mpoPlane->mpoMap->nGetHeight();
	
	//read in obstacle list
	long plListSize;
	long *plObstList;
	
	//update the exterior obstacles first
	plObstList = mpoTemplate->plGetList(TAG_Lobs, &plListSize);
	MINI_MAP_TERRAIN_TYPE eMMTType = (MINI_MAP_TERRAIN_TYPE)mpoTemplate->lGetValue(TAG_mmtt, 0);
	if (plListSize) {
		// we have data in Lobs tag
		if (plListSize % 2 == 0) {	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) {
				nNewNodeX = nXPos + plObstList[i];
				nNewNodeY = nYPos + plObstList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) {
					MapNode *poNode = mpoPlane->mpoMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNode);
					//Any drawobj can have a Lobs list now (like grass) so determine which ones mark node as an obstacle
					switch (eMMTType)
					{
						case MMTT_NONE:
							poNode->meType = TT_OBSTACLE;
							break;
						case MMTT_GRASS:
							if (poNode->meMMType == MMTT_NONE)
							{
								//only mark if not previously marked 
								poNode->meMMType = eMMTType;
							}
							break;
						case MMTT_SAND:
							if (poNode->meMMType == MMTT_NONE)
							{
								//only mark if not previously marked 
								poNode->meMMType = eMMTType;
							}
							break;
						case MMTT_SNOW:
							if (poNode->meMMType == MMTT_NONE)
							{
								//only mark if not previously marked 
								poNode->meMMType = eMMTType;
							}
							break;
						case MMTT_OBSTACLE:
							poNode->meType = TT_OBSTACLE;
							poNode->meMMType = eMMTType;

							break;
						case MMTT_BUILDING:
							poNode->meType = TT_OBSTACLE;
							poNode->meMMType = eMMTType;
							break;
						case MMTT_TREES:
							poNode->meType = TT_OBSTACLE;
							poNode->meMMType = eMMTType;
							break;
						case MMTT_STONE:
							poNode->meType = TT_OBSTACLE;
							poNode->meMMType = eMMTType;
							break;
						case MMTT_ROADS:
							poNode->meMMType = eMMTType;
							break;
						case MMTT_WATER:
							poNode->meType = TT_OBSTACLE;
							poNode->meMMType = eMMTType;
							break;
						case MMTT_BRIDGE:
							poNode->meType = TT_OBSTACLE; //TBD - open/closed??
							poNode->meMMType = eMMTType;
							break;
						case MMTT_GATE:
							poNode->meType = TT_OBSTACLE; //TBD - open/closed??
							poNode->meMMType = eMMTType;
							break;
						case MMTT_DIRT:
							if (poNode->meMMType == MMTT_NONE)
							{
								//only mark if not previously marked 
								poNode->meMMType = eMMTType;
							}
							break;
					}
					if (poNode->meType == TT_OBSTACLE)
					{
						poNode->vAddCover(this);
					}
				}
			}
		}
		else {
			//if there's 0 or 1 number after tag, plListSize is always 1, so just ignore the
			//list if it reports a size of 1, else odd numbered lists 3 and above will get the
			//debug message below
			if (plListSize > 2)	//ignore this message for "Lobs=" or "Lobs=x" lines (x could be any number by itself)
				vDebugOut ("Odd numbered obstacle list \n");
		}
	}

	//update the interiors next
	long *plIntList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize) {
		// we have data in Libs tag
		if (plListSize % 2 == 0) {	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) {
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) {
					MapNode *poNode = mpoPlane->mpoMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNode);
					poNode->meType = TT_INTERIOR;
					poNode->meMMType = MMTT_BUILDING; //only buildings use the Libs so its OK to assign this here
					poNode->vAddCover(this);
				}
			}
		}
		else {
			//if there's 0 or 1 number after tag, plListSize is always 1, so just ignore the
			//list if it reports a size of 1, else odd numbered lists 3 and above will get the
			//debug message below
			if (plListSize > 2)	//ignore this message for "Lobs=" or "Lobs=x" lines (x could be any number by itself)
				vDebugOut ("Odd numbered interior list %d\n");
		}
	}

}

void DrawObj::vClearObstacles(GameMap *poMap, int nWidth, int nHeight) {
	//read in obstacle list
	long plListSize;
	long *plObstList;
	
	//clear the exterior obstacles first
	plObstList = mpoTemplate->plGetList(TAG_Lobs, &plListSize);
	if (plListSize) {
		// we have data in Lobs tag
		if (plListSize % 2 == 0) {	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) {
				nNewNodeX = nXPos + plObstList[i];
				nNewNodeY = nYPos + plObstList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) {
					MapNode *poNode = poMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNode);
					poNode->meType = TT_CLEAR;
					poNode->meMMType =  MMTT_GRASS;
					poNode->vClearCover();
				}
			}
		}
	}

	
	//clear the interior list next
	long *plIntList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize) {
		// we have data in Libs tag
		if (plListSize % 2 == 0) {	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) {
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) {
					MapNode *poNode = poMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNode);
					poNode->meType = TT_CLEAR;
					poNode->meMMType =  MMTT_GRASS;
					poNode->vClearCover();
				}
			}
		}
	}
}

void DrawObj::vMapEditHandleEvent(sciEvent &oEvent) {
//vDebugOut ("type %d mod:%d\n", oEvent.type,oEvent.modifiers);
	
	//Before we can handle the various editing events, we need to find out what
	//editing mode we're in

	switch (gpoMapEdit->eMode()) {
	
	// In mode EM_NONE:
	//		LMB - selects and displays data on map edit plane
	//		RMB - sets mode to EM_DRAG and makes this object an event hog
	//		'C' - cycle "C"el by one
	//		'G' - enter obstacle "G"rid mode
	//		'L'		   - cycle "L"oop by one
	//		'P' - enter orgin "P"oint mode
	case EM_NONE:
		switch (oEvent.type) {
		case MOUSE_DOWN:
			if (oEvent.modifiers & LEFT_MOUSE) 
			{
				if (bOnMe(oEvent)) { 
					if (mbSelectable) {
//						mpoPlane->mpoMap->vClearAllSelections();
//						mpoPlane->mpoMap->vAddSelection(this);
//						gpoMapEdit->vSetMode(EM_NONE);
						gpoMapEdit->vSetObject(this);
						gpoMapEdit->vMove(this);	//either select or move self
						vMarkUpdate();
						oEvent.claimed = TRUE;
					} else {
						//we clicked on a non selectable thing (like grass)
						// TBD - Edit grass. Maybe layers. For now just clear selections
						mpoPlane->mpoMap->vClearAllSelections();
						gpoMapEdit->vSetObject(NULL);
						vMarkUpdate();
						oEvent.claimed = TRUE;
					}
				} //OnMe
			}
			else if (oEvent.modifiers & RIGHT_MOUSE)
			{
// tw- commented out since Don wants a left click to drag, plus I got rid of the way of dragging. Have to go through AddPlane to drag around
/*				if (bOnMe(oEvent) && mbSelectable) { //TBD for now only drag Selectable things, ie NO GROUND
					gpoMapEdit->vSetMode(EM_DRAG);
					gpoMapEdit->vSetObject(this);
					gpoMapEdit->vSetUnitDrag(TRUE);
					if (bIsBuilding()) {
						//read in obstacle lists
						gpoMapEdit->vClearObstLists();
						vReadInLists();

						//read in interior unit list
						gpoMapEdit->vSetLists((Building*)this);
					}
					mpoPlane->mpoMap->vClearAllSelections();
					assert((DrawObj*)this);
					// place mouse cursor on node where origin point sits
					POINT pt1 = mpoPlane->mpoMap->oCalcCoords(mpoCurNode->mnXPos, mpoCurNode->mnYPos);
					ClientToScreen(pApp->pWindow->hWnd, &pt1);
					SetCursorPos (pt1.x, pt1.y);
					//vDragObj((DrawObj*)this); //the old procedure call
					pApp->pWindow->vAddEventHog(this);
					oEvent.claimed = TRUE;
				}
*/
			} //oEvent.modifiers (EM_NONE){
			break;	// MOUSE_DOWN(EM_NONE)
			case KEY_PRESS:
				if (mbSelected && mpoPlane->mpoMap->nSelectionSize() == 1) {
					switch (oEvent.c) {

					//step through each cel
					case 'C':
						if (oEvent.modifiers & SHIFT_DOWN)
							vDecrementCel();
						else
							vIncrementCel();
						break;

/*					//perform a forwards cycle
					case 'F':
						if (mbDoneCycling)
							mpoCycler = new ForwardCycler(this,mpoBitmap);
						oEvent.claimed = TRUE;
						break;
*/
					//Edit obstacle list 'G'rids (Lobs or Libs)
					case 'G':
						//no reason to edit obstacle lists for units
						if (meClass != OC_UNIT) {
							gpoMapEdit->vSetMode(EM_OBSTACLE);
							gpoMapEdit->vSetObstMode(V_EXTERIOR);
							gbToggleObstacleGrid = FALSE;
							gpoMapEdit->vSetObject(this);
							mpoPlane->mpoMap->vClearAllSelections();
							gpoMapEdit->vClearObstLists();
							vReadInList(TAG_Lobs, &(gpoMapEdit->moExtPointList));
							vReadInList(TAG_Libs, &(gpoMapEdit->moIntPointList));
							//vDrawObstacles();
							gpoMapEdit->vMarkUpdate();
							pApp->pWindow->vAddEventHog(this);
							oEvent.claimed = TRUE;
						}
						break; // 'G'

					case 'L':
						vSetCel(0);
						if (oEvent.modifiers & SHIFT_DOWN)
							vDecrementLoop();
							//mpoBitmap->SetLoop(mpoBitmap->nGetLoop() - 1);
						else
							vIncrementLoop();
							//mpoBitmap->SetLoop(mpoBitmap->nGetLoop() + 1);
						break;


					//Edit origin 'P'oint 
					case 'P':
						gpoMapEdit->vSetMode(EM_ORIGIN);
						gbToggleObstacleGrid = FALSE;
						gpoMapEdit->vSetObject(this);
//						mpoPlane->mpoMap->vClearAllSelections();
						gpoMapEdit->vSetOriginStart(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
						pApp->pWindow->vAddEventHog(this);
						oEvent.claimed = TRUE;
						break; // 'P'

/*					//perform a reverse cycle
					case 'R':
						if (mbDoneCycling)
							mpoCycler = new ReverseCycler(this,mpoBitmap);
						oEvent.claimed = TRUE;
						break;
*/
					} // oEvent.c
				} // (mbSelected && nSelectionSize)
				break; // KEY_PRESS(EM_NONE)
		} // oEvent.type(EM_NONE)
		break; //EM_NONE


	// In mode EM_DRAG:
	//		UP   - a mouse up (either left or right) stops drag mode and sets mode to EM_NONE
	//		MOVE - moves object to node the cursor is over until a MOUSE_UP is detected
//tw - Dragging functionality moved to AddPlane
/*	case EM_DRAG:
		switch (oEvent.type) {
		case MOUSE_UP:
			gpoMapEdit->vSetMode(EM_NONE);
			gpoMapEdit->vSetObject(NULL);
			pApp->pWindow->vRemoveEventHog(this);
			if (bIsBuilding()) {
				gpoMapEdit->vUpdateUnitPositions((Building*)this);
			}
			mpoPlane->mpoMap->vUpdateObstacles();
			vMarkUpdate();
			oEvent.claimed = TRUE;
			break; // MOUSE_UP(EM_DRAG)
		case MOUSE_MOVE:
			oEvent.vLocalize(mpoPlane);	//need to localize it because its a hog 
			vDrag(oEvent);
			oEvent.claimed = TRUE;	//we're 'hogging' events until we release the button
			break; // MOUSE_MOVE(EM_DRAG)
		} // oEvent.type(EM_DRAG)
		break; //EM_DRAG
*/

	// In mode EM_ADD:
	//		LMB  - single add at node and stops add mode and sets to EM_NONE
	//		RMB  - multiple adds and remains in mode
	//		ESC  - exits mode and sets to EM_NONE
	//		MOVE - moves object to node the cursor is over until we exit mode
// tw - Adding functionality moved to AddPlane
/*	case EM_ADD:
		switch (oEvent.type) {
		case MOUSE_DOWN:
			oEvent.vLocalize(mpoPlane);	//need to localize it because its a hog 
			if (oEvent.modifiers & LEFT_MOUSE) 
			{
				pApp->pWindow->vRemoveEventHog(this);
				vInsertSelf(oEvent);
				gpoMapEdit->vSetMode(EM_NONE); //this lines NEEDS to come after vInsertSelf
				gpoMapEdit->vSetObject(NULL);
				//mpoPlane->mpoMap->vDraw(); //can't use this since there's a possibility of being deleted, moved below
			}
			else if (oEvent.modifiers & RIGHT_MOUSE) 
			{
				vCopySelf(oEvent);
			} // oEvent.modifiers
			gpoMap->vUpdateObstacles(); // need to use global 'gpo' since self may be deleted at this point
			gpoMapEdit->vMarkUpdate();
			oEvent.claimed = TRUE;
			break; // MOUSE_DOWN
		case MOUSE_MOVE:
			oEvent.vLocalize(mpoPlane);	//need to localize it because its a hog 
			vDrag(oEvent);
			oEvent.claimed = TRUE;	//we're 'hogging' events until we release the button
			break; // MOUSE_MOVE
		case KEY_PRESS:
			switch (oEvent.c) {
			case VK_ESCAPE:	// 'ESC' key
				pApp->pWindow->vRemoveEventHog(this);
				//this->mpoCurNode->vRemoveObj(this); //this line not needed now that i put it in ~DrawObj
				delete this;
				gpoMapEdit->vSetMode(EM_NONE);
				gpoMapEdit->vSetObject(NULL);
				gpoMapEdit->vMarkUpdate();
				//gpoMap->vDraw();
				break;
			}
			break; // KEY_PRESS
		} // oEvent.type

		oEvent.claimed = TRUE;
		break; //EM_ADD
*/

	// In mode EM_ORIGIN:
	//		Arrow keys - move(and sets) origin point by 1
	//		SHIFT      - uses above keys and changes step to 5 instead of 1
	//		ESC        - aborts origin point change and returns to starting position
	//		ENTER      - writes out new TEX file with changed origin point
	//		'C'		   - cycle "C"el by one
	//		'L'		   - cycle "L"oop by one
	//		'X'		   - center origin point "X"wise
	//		'Y'		   - center origin point "Y"wise
	case EM_ORIGIN:
		oEvent.vLocalize(mpoPlane);	//need to localize it because its a hog 
		switch (oEvent.type) {
		case KEY_PRESS:
			{
				short nStep = 1;
				if (oEvent.modifiers & SHIFT_DOWN) {
					nStep = 10;//5;
				}
				
				switch (oEvent.c) {
				case VK_ESCAPE:
					//reselect
//					mpoPlane->mpoMap->vClearAllSelections();
//					mpoPlane->mpoMap->vAddSelection(gpoMapEdit->poObject());
//					mpoPlane->mpoMap->vCommitSelection();
					{
						//mpoBitmap->vSetOriginPoint(gpoMapEdit->nStartX(), gpoMapEdit->nStartY());
						mpoBitmap->vMoveAllOrigins(gpoMapEdit->nStartX() - mpoBitmap->nOriginX, gpoMapEdit->nStartY() - mpoBitmap->nOriginY); 
						gpoMapEdit->vMarkUpdate();
						pApp->pWindow->vRemoveEventHog(this);
						gpoMapEdit->vSetMode(EM_NONE);
						gpoMapEdit->vSetObject(NULL);
						break; //VK_ESCAPE
					}
				
				case VK_RETURN:
					vSaveBitmap();
					gpoMapEdit->vMarkUpdate();
					pApp->pWindow->vRemoveEventHog(this);
					mpoPlane->mpoMap->vUpdateObstacles();
					gpoMapEdit->vSetMode(EM_NONE);
					gpoMapEdit->vSetObject(NULL);
					break; // VK_RETURN

				case VK_RIGHT:
//					if ( mpoBitmap->nOriginX + nStep < mpoBitmap->Width() ) {
						//mpoBitmap->vSetOriginPoint(mpoBitmap->nOriginX + nStep, mpoBitmap->nOriginY);
						mpoBitmap->vMoveAllOrigins(nStep, 0);
						gpoMapEdit->vSetCurrent(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
//					}
					oEvent.claimed = TRUE;
					break; // VK_RIGHT
				case VK_LEFT:
//					if ( mpoBitmap->nOriginX - nStep >= 0 ) {
						//mpoBitmap->vSetOriginPoint(mpoBitmap->nOriginX - nStep, mpoBitmap->nOriginY);
						mpoBitmap->vMoveAllOrigins(-nStep, 0);
						gpoMapEdit->vSetCurrent(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
//					}
					oEvent.claimed = TRUE;
					break; // VK_LEFT
				case VK_DOWN:
//					if ( mpoBitmap->nOriginY + nStep < mpoBitmap->Height() ) {
						//mpoBitmap->vSetOriginPoint(mpoBitmap->nOriginX, mpoBitmap->nOriginY + nStep);
						mpoBitmap->vMoveAllOrigins(0, nStep);
						gpoMapEdit->vSetCurrent(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
//					}
					oEvent.claimed = TRUE;
					break; // VK_DOWN
				case VK_UP:
//					if ( mpoBitmap->nOriginY - nStep >= 0) {
						//mpoBitmap->vSetOriginPoint(mpoBitmap->nOriginX, mpoBitmap->nOriginY - nStep);
						mpoBitmap->vMoveAllOrigins(0, -nStep);
						gpoMapEdit->vSetCurrent(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
//					}
					oEvent.claimed = TRUE;
					break; // VK_UP
				case 'C':
					if (oEvent.modifiers & SHIFT_DOWN)
						vDecrementCel();
					else
						vIncrementCel();
					break;

				case 'L':
					vSetCel(0);
					if (oEvent.modifiers & SHIFT_DOWN)
						vDecrementLoop();
						//mpoBitmap->SetLoop(mpoBitmap->nGetLoop() - 1);
					else
						vIncrementLoop();
						//mpoBitmap->SetLoop(mpoBitmap->nGetLoop() + 1);
					break;

				case 'X':
					{
						int nOffset = (mpoBitmap->Width() / 2 - mpoBitmap->nOriginX);
						//mpoBitmap->vSetOriginPoint(mpoBitmap->Width() / 2, mpoBitmap->nOriginY);
						mpoBitmap->vMoveAllOrigins(nOffset, 0);
						gpoMapEdit->vSetCurrent(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
						oEvent.claimed = TRUE;
						break; // 'X'
					}
				case 'Y':
					{
						int nOffset = (mpoBitmap->Height() / 2 - mpoBitmap->nOriginY);
						//mpoBitmap->vSetOriginPoint(mpoBitmap->nOriginX, mpoBitmap->Height() / 2);
						mpoBitmap->vMoveAllOrigins(0, nOffset);
						gpoMapEdit->vSetCurrent(mpoBitmap->nOriginX, mpoBitmap->nOriginY);
						gpoMapEdit->vMarkUpdate();
						oEvent.claimed = TRUE;
						break; // 'Y'
					}

					
				} //oEvent.c

				break; // KEY_PRESS
			}

		} // oEvent.type
		oEvent.claimed = TRUE;	//always claim as TRUE since we're really a MODAL hog
		break; //EM_ORIGIN


	// In mode EM_OBSTACLE:
	//		LMB   - toggle an obstacle grid for the 'Lobs' list
	//		RMB   - toggle an obstacle grid for the 'Libs' list for buildings with Interiors ONLY
	//		ENTER - writes out the lists if they've changed in any way
	//		V     - toggles interior/exterior views for buildings
	//		ESC   - Cancels any changes made to either list
	case EM_OBSTACLE:
		oEvent.vLocalize(mpoPlane);	//need to localize it because its a hog 
		switch (oEvent.type) {
		case KEY_PRESS:
			switch (oEvent.c) {
			case VK_RETURN:
				//check to see if list has changed
				if (gpoMapEdit->mbListChanged) {
					vWriteList(TAG_Lobs, &(gpoMapEdit->moExtPointList));
					if (bIsBuilding && nInsideBitmapID()) {
						//if we're a building and we have an interior then write Libs list
						vWriteList(TAG_Libs, &(gpoMapEdit->moIntPointList));
					}
					mpoPlane->mpoMap->vUpdateObstacles();
				}
				//note there is no break here
			case VK_ESCAPE:
				//reselect
				mpoPlane->mpoMap->vClearAllSelections();
				mpoPlane->mpoMap->vAddSelection(gpoMapEdit->poObject());
				mpoPlane->mpoMap->vCommitSelection();
				gpoMapEdit->vMarkUpdate();
				pApp->pWindow->vRemoveEventHog(this);
				gpoMapEdit->vSetMode(EM_NONE);
				gpoMapEdit->vSetObject(NULL);
				break; //VK_ESCAPE
			case VK_RIGHT:
				gpoMapEdit->vXListScroll(1);
				gpoMapEdit->vMarkUpdate();
				oEvent.claimed = TRUE;
				break; // VK_RIGHT
			case VK_LEFT:
				gpoMapEdit->vXListScroll(-1);
				gpoMapEdit->vMarkUpdate();
				oEvent.claimed = TRUE;
				break; // VK_LEFT
			case VK_UP:
				gpoMapEdit->vYListScroll(-1);
				gpoMapEdit->vMarkUpdate();
				oEvent.claimed = TRUE;
				break; // VK_UP
			case VK_DOWN:
				gpoMapEdit->vYListScroll(1);
				gpoMapEdit->vMarkUpdate();
				oEvent.claimed = TRUE;
				break; // VK_DOWN

			case 'V': 
				gbToggleInteriors = abs(gbToggleInteriors - 1);
				mpoPlane->mpoMap->vDraw();
				break; // 'V'

			} // oEvent.c
			break; // KEY_PRESS(EM_OBSTACLE)
		case MOUSE_DOWN:
			if (bIsBuilding() && nInsideBitmapID()) {
			//if we're a building and we have an interior, then allow the right mouse button
			//to toggle the interior 'Libs' list
				if (oEvent.modifiers & RIGHT_MOUSE) {
					gpoMapEdit->vSetObstMode(V_INTERIOR);
				} else {
					gpoMapEdit->vSetObstMode(V_EXTERIOR);
				}
			}
			gpoMapEdit->vToggleObstacle(oEvent);
			gpoMapEdit->vMarkUpdate();
			break; // MOUSE_DOWN(EM_OBSTACLE)
		} // oEvent.type(EM_OBSTACLE)
		
		oEvent.claimed = TRUE;	//always claim as TRUE since we're really a MODAL hog
		break; //EM_OBSTACLE

	}	//switch (gpoMapEdit->eMode())

}

void DrawObj::vDrag(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		if (mpoCurNode != poNode) {
			
			// The mouse is in a different node than the current object

			//mpoCurNode->vRemoveObj(this); //no need for this line, see below
			// Note - unlike Unit::vDrag, there's no need to put back on clipboard if we were on
			// it. I allow multiple copies of a drawobj to go into a node.

			poNode->vAddObj(this); //this lines also removes self from previous node
			gpoMap->vUpdateObstacles();
			oEvent.claimed = TRUE;	
			vMarkUpdate();
		}
	}
}

// For adding to the map.
// The draw object being dragged around makes a copy of itself and inserts in into the
// node nearest the event
void DrawObj::vCopySelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {
		DrawObj *poCopy = new DrawObj(mpoPlane, nTemplateID());	
		poCopy->vSetLayer(mnLayer);
		poNode->vAddObj(poCopy);
	}
}

// For adding to the map.
// The draw object being dragged around inserts ITSELF in the node nearest the event.
void DrawObj::vInsertSelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		//most likely vDrag will have already inserted us into the node(and into moDrawn)
		//but if it didn't we need to catch it here.
//		vSetLayer(DP_OBSTACLES);
		poNode->vAddObj(this); //this lines also removes self from previous node
		//gpoMapEdit->mpoMapNode->vReset(); //clear out clipboard - no need see above line

	} else {
		// we clicked somewhere not on the game map so get rid of self
		delete this;
	}
}

Boolean DrawObj::HandleEvent(sciEvent & oEvent)
{
	if (gbMapEditing) 
	{
		assert(gpoMapEdit);				//if this is ever NULL here, we're in trouble

		// map editing operation
		vMapEditHandleEvent(oEvent);	//passing to another handler to keep this one cleaner		
		return oEvent.claimed;
	}

	if (mpoParent && mbEventBubbling && bOnMe(oEvent))
		return (mpoParent->HandleEvent(oEvent));
	else
		return oEvent.claimed;
}


void DrawObj::vDraw(int nX, int nY) 
{
	assert(mpoBitmap);
	if (!mbVisible) {
		return;
	}
	mpoBitmap->SetLoop(mnLoop);
	mpoBitmap->SetCel(mnCel);
	mpoBitmap->vSetFlip(mbFlip);
	mpoBitmap->vDraw(nX, nY, mpoPlane, mjAlpha, 0, 0, &mrCurrent, mpwSpecialPalette);
	mpoPlane->vAddToDrawnList(this);
}


void DrawObj::vDraw(void)
{
	assert(mpoCurNode);
	if (!mbVisible) {
		return;
	}

	if (mpoParent)// && mpoParent->bIsBuilding())
	{
		if (mbParentOpen)//(mpoParent->bIsOpen()) 
		{	
			if (meDrawState == DS_EXTERIROR)		//if we're an exterior only child
				return;							//then don't draw since we're looking at the interior
		} else { 
			if (meDrawState == DS_INTERIOR)		//if we're an interior only child
				return;							//then don't draw since we're looking at the exterior
		}
	}

	assert(mpoBitmap);
	mpoBitmap->SetLoop(mnLoop);
	mpoBitmap->SetCel(mnCel);
	mpoBitmap->vSetFlip(mbFlip);

	if (mbSelected) 
	{
		vHighlight();
	}

	mpoBitmap->vDraw(nPixelX() - mpoPlane->nMapOrgPixelX(), nPixelY() - mpoPlane->nMapOrgPixelY(), mpoPlane, mjAlpha, 0, 0, &mrCurrent, mpwSpecialPalette);

	// this whole thing should probably be replaced
	// with a call to the bitmap to get the current rect
	// TBD -- DJM
	/*
	int nWidth = mpoBitmap->Width();
	int nHeight = mpoBitmap->Height();

	nX += -(mpoBitmap->nOriginX);
	nY += -(mpoBitmap->nOriginY);

	SetRect(&mrCurrent, nX, nY, nX + nWidth, nY + nHeight);
	*/

	// I'm drawing, so add myself to the drawn list
	mpoPlane->vAddToDrawnList(this);

}

void DrawObj::vHighlight() { 
	//RECT *prClipRect;
	WORD wObjWidth =  mpoBitmap->Width();
	WORD wObjHeight = mpoBitmap->Height();

	int tmpTop    = mrCurrent.top; 
	int tmpLeft   = mrCurrent.left;
	mpoPlane->vLocalToGlobal(&tmpLeft, &tmpTop);
	int tmpRight = (tmpLeft + wObjWidth);
   int tmpBottom = (tmpTop + wObjHeight);

	RECT rLocal = {mrCurrent.left, mrCurrent.top, mrCurrent.left + wObjWidth, mrCurrent.top + wObjHeight};
	
/*	
	//The 'Pen' style of highlighting
	HDC pDC = pApp->pWindow->hDC;
	prClipRect = mpoPlane->prGetClip();
		
	HRGN hClipRegion = CreateRectRgn(prClipRect->left, prClipRect->top, prClipRect->right, prClipRect->bottom);
	SelectClipRgn(pDC, hClipRegion);
	
	HGDIOBJ hThePen;
	hThePen = CreatePen(PS_SOLID, 2, 0x0000ffff); //0x00bbggrr - if we want different colors, use this
	//hThePen = GetStockObject(WHITE_PEN); 	// if stock white pen is fine, then use this
	SelectObject (pDC, hThePen);
   
	MoveToEx (pDC, tmpLeft, tmpTop, NULL);
	LineTo (pDC, tmpRight, tmpTop);
	LineTo (pDC, tmpRight, tmpBottom);
	LineTo (pDC, tmpLeft, tmpBottom);
	LineTo (pDC, tmpLeft, tmpTop);
	DeleteObject(hThePen);
*/	

	if (bIsUnit())
	{
		Bitmap * pBM = GetHighlightBitmap();
		if (pBM)
		{
			static mnCel = 0;
			static snDirection = 1;
			static int snCycle = 0;
			static snOffset = 0;
			int nCycle = game->nGameCycle();
			int nMax = pBM->GetNumCels()-1;
			if (nCycle != snCycle)
			{
				snCycle = nCycle;
				
				if (mnCel == nMax)
				{
					snDirection = -1;
				}			
				else if (mnCel == 0)
				{
					snDirection = 1;
				}
				mnCel += snDirection;
				snOffset = 0;
			}
			else
			{
				snOffset++;
			}
			pBM->SetCel((mnCel + snOffset) % nMax);
			pBM->vDraw(nPixelX() - mpoPlane->nMapOrgPixelX(), nPixelY() - mpoPlane->nMapOrgPixelY(), mpoPlane, 96);
			return;
		}
		// FALLTHROUGH
	}


	Bitmap *pBM = new Bitmap(wObjWidth, wObjHeight);
	pBM->DrawRect(game->ColorHighlight(), rLocal, gpoMap->mpoPlane, 1);
	delete pBM;
}

void DrawObj::vMarkUpdate(void)
{
	// for now, mark everything
	mpoPlane->window->Update(*(mpoPlane->prGetClip()));
}

void DrawObj::vSaveMap(FILE *file, WORD wVersion) {
	int nResID = mpoTemplate->resourceID;
	int nLayer = mnLayer;

	switch (wVersion) {
	case 1:
		fwrite(&nResID, sizeof(int), 1, file);	//write the resourceID of the template
		fwrite(&nLayer, sizeof(int), 1, file); //write the draw layer the object is in
		break;

	}
}

void DrawObj::vLoadMap(FILE *file, WORD wVersion) 
{

}

void DrawObj::vReadInList(long lTag, List<ObstaclePoint> *poPointList) {
	//initialze PointLists based on current template tag(lTag) and put into appropiate list(poPointList).

	long plListSize;
	long *plObstList;
	int i;

	plObstList = mpoTemplate->plGetList(lTag, &plListSize);
	if (plListSize % 2 == 0) {
		// we have data in tag thats an even multiple of 2
		for (i=0; i < plListSize; i+=2) {
			ObstaclePoint *poPnt = new ObstaclePoint(plObstList[i], plObstList[i+1]);
			poPointList->AddToEnd(poPnt);
		}
	}
	
	//interior obstalces
/*	plObstList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize % 2 == 0) {
		// we have data in Libs tag thats an even multiple of 2
		for (i=0; i < plListSize; i+=2) {
			ObstaclePoint *poPnt = new ObstaclePoint(plObstList[i], plObstList[i+1]);
			gpoMapEdit->moIntPointList.AddToEnd(poPnt);
		}
	}
*/}


void DrawObj::vReadInList(long lTag, List<IntegerPoint> *poPointList) {
	//initialze lists based on current template tag(lTag) and put into appropiate list(poPointList).

	long plListSize;
	long *plIntList;
	int i;

	plIntList = mpoTemplate->plGetList(lTag, &plListSize);
	if (plListSize) {
		// we have data in tag
		for (i=0; i < plListSize; ++i) {
			IntegerPoint *poPnt = new IntegerPoint(plIntList[i]);
			poPointList->AddToEnd(poPnt);
		}
	}
}
	
	
void DrawObj::vDrawObstacles() {
	int nSize, i;
	int nWidth  =  mpoPlane->mpoMap->nGetSpacingX();
	int nHeight =  mpoPlane->mpoMap->nGetSpacingY();
//	int nOriginX = mpoBitmap->nOriginX;
//	int nOriginY = mpoBitmap->nOriginY;
	int nBaseX = nPixelX() - mpoPlane->nMapOrgPixelX();
	int nBaseY = nPixelY() - mpoPlane->nMapOrgPixelY();	//

	//center the rectangle 'around' the node point
	int nX = ( nBaseX - (nWidth / 2) );
	int nY = ( nBaseY - (nHeight /2) );
	RECT rLocal = {nX, nY, nX+nWidth, nY+nHeight};
	Bitmap *pBM = new Bitmap(nWidth, nHeight);
	pBM->DrawRect(C_WHITE, rLocal, mpoPlane, 2); 
	delete pBM;

	//draw the exterior obstacles first
	nSize = gpoMapEdit->moExtPointList.GetSize();
	for (i = 0; i < nSize; ++i) {
		ObstaclePoint *poPoint = gpoMapEdit->moExtPointList.GetItem(i);
		nX = ( (poPoint->mnXPos * nWidth) +  nBaseX - (nWidth / 2) );
		nY = ( (poPoint->mnYPos * nHeight) + nBaseY - (nHeight /2) );
		SetRect (&rLocal, nX, nY, nX+nWidth, nY+nHeight);
		Bitmap *pBM = new Bitmap(nWidth, nHeight);
		pBM->DrawRect(C_OBSTACLE, rLocal, mpoPlane, 1); 
		delete pBM;
	}

	
	//then draw the interiors
	nSize = gpoMapEdit->moIntPointList.GetSize();
	for (i = 0; i < nSize; ++i) {
		ObstaclePoint *poPoint = gpoMapEdit->moIntPointList.GetItem(i);
		ObstaclePoint *poDupPnt = gpoMapEdit->poFindPoint(poPoint->mnXPos, poPoint->mnYPos, &(gpoMapEdit->moExtPointList));
		//check for duplicate point in exterior list, if so, delete from interior list
		if (poDupPnt != NULL) {
			vDebugOut ("Duplicate point x:%d,y:%d found in 'Libs'\n", poPoint->mnXPos, poPoint->mnYPos);
			gpoMapEdit->moIntPointList.Remove(poPoint);
			delete poPoint;
			i = 0; //scan the list from the beginning again
			nSize = gpoMapEdit->moIntPointList.GetSize();
		}
		else {
			//neither list has the same point, so draw interir point
			nX = ( (poPoint->mnXPos * nWidth) +  nBaseX - (nWidth / 2) );
			nY = ( (poPoint->mnYPos * nHeight) + nBaseY - (nHeight /2) );
			SetRect (&rLocal, nX, nY, nX+nWidth, nY+nHeight);
			Bitmap *pBM = new Bitmap(nWidth, nHeight);
			pBM->DrawRect(C_INTERIOR, rLocal, mpoPlane, 1); 
			delete pBM;
		}
	}

}

void DrawObj::vWriteList(long lTag, List<ObstaclePoint> *poPointList) {
	char str[5]; 
	memset(str, 0x0, 5);
	*(ULONG*)str= lTag;

	char strFileName[64];
	sprintf(strFileName, "%s\\%d.%s", ".", nTemplateID(), "tlt");

	char strScratch[12]; //enough room for xxxxx,yyyyy plus a null terminator
	char strList [1024];
	memset(strList, 0x0, 1024);
	
	int nSize, i;
	nSize = poPointList->GetSize();
	for (i = 0; i < nSize; ++i) {
		ObstaclePoint *poPoint = poPointList->GetItem(i);
		//the 'x'
		memset(strScratch, 0x0, 12);
		sprintf(strScratch, "%d,", poPoint->mnXPos);
		strcat(strList, strScratch);
		//the 'y'
		memset(strScratch, 0x0, 12);
		if (i + 1 == nSize) {
			sprintf(strScratch, "%d", poPoint->mnYPos);
		}
		else {
			sprintf(strScratch, "%d, ", poPoint->mnYPos);
		}
		strcat(strList, strScratch);
	}
	
	if (nSize) {
		//there is at least 1 pair of obstacle coords to write out
		WritePrivateProfileString("lists", str, strList, strFileName); 
	}
	else {
		//no data so delete tag from .tlt file
		WritePrivateProfileString("lists", str, NULL, strFileName); 
	}

	//flush the cache
	WritePrivateProfileString(NULL,NULL, NULL, strFileName);

	//reload the changed data
	mpoTemplate->vReload();

	//update the new obstacle list
	mpoPlane->mpoMap->vUpdateObstacles();
	
}

void DrawObj::vWriteList(long lTag, List<IntegerPoint> *poPointList) {
	char str[5]; 
	memset(str, 0x0, 5);
	*(ULONG*)str= lTag;

	char strFileName[64];
	sprintf(strFileName, "%s\\%d.%s", ".", nTemplateID(), "tlt");

	char strScratch[12]; //enough room for xxxxx,yyyyy plus a null terminator
	char strList [1024];
	memset(strList, 0x0, 1024);
	
	int nSize, i;
	nSize = poPointList->GetSize();
	for (i = 0; i < nSize; ++i) {
		IntegerPoint *poPoint = poPointList->GetItem(i);
		//the 'integer'
		memset(strScratch, 0x0, 12);
		if (i + 1 == nSize) {
			sprintf(strScratch, "%d", poPoint->mnNumber);
		}
		else {
			sprintf(strScratch, "%d, ", poPoint->mnNumber);
		}
		strcat(strList, strScratch);
	}
	
	if (nSize) {
		//there is at least 1 integer to write out
		WritePrivateProfileString("lists", str, strList, strFileName); 
	}
	else {
		//no data so delete tag from .tlt file
		WritePrivateProfileString("lists", str, NULL, strFileName); 
	}

	//flush the cache
	WritePrivateProfileString(NULL,NULL, NULL, strFileName);

	//reload the changed data
	mpoTemplate->vReload();

	//refresh obstacle list
	mpoPlane->mpoMap->vUpdateObstacles();

	//refresh map
	mpoPlane->mpoMap->vDraw();
	
}



void DrawObj::vSaveBitmap(void) {
	RESOURCE_TYPES oldType;

	//need to 'fake' the resource into thinking it's a ViewBitmap
	//so it'll attach TEX as an extension instead of BMP
	oldType = mpoBitmap->type;
	mpoBitmap->type = RES_VIEWBITMAP;
										
	//write standard view 
	mpoBitmap->vWriteFile();
	
	//restore to original just in case
	mpoBitmap->type = oldType;
	
}

void DrawObj::GetPlayPosition(unsigned int * unX, unsigned int *unY)
{
	*unX = nPixelX();
	*unY = nPixelY();
}

void DrawObj::vChangeToInside() {
}
  
void DrawObj::vChangeToOutside() {
}

void DrawObj::vSetForcing(Boolean bForce) {
}

/*void DrawObj::vGameCycle(void)
{
	//vAnimationCycle(); //TBD - for now pass it to the animation cycle may need animation cycle to occur faster than game cycles
}

void DrawObj::vAnimationCycle(void)
{
	if (!mbDoneCycling && mpoCycler)
	{
		mpoCycler->vAnimationCycle();
	}
}
*/
void DrawObj::vCue()
{
/*	if (gbMapEditing) 
	{
		delete mpoCycler;
		mpoCycler = NULL;
//		mpoBitmap->SetCel(0);
//		vMarkUpdate();
		mpoPlane->mpoMap->vClearAllSelections();
		mpoPlane->mpoMap->vAddSelection(this);
	}
*/
}

void DrawObj::vSetParent(DrawObj *poDrawObj)
{
	mpoParent = poDrawObj;
}


void DrawObj::vRemoveChild(DrawObj *poObj)
{
	//if (mpoChildren)
	if (bChildren())
	{
		mpoChildren->Remove(poObj);
//		if (mpoChildren->GetSize() == 0)
//		{
//			delete mpoChildren;
//			mpoChildren = NULL;
//		}
	}
}

void DrawObj::vInitChildren()
{
	//read in list
	long plListSize;
	long *plChildList;
	
	plChildList = mpoTemplate->plGetList(TAG_ctem, &plListSize);
	if (plListSize) {
		mpoChildren = new RefList<DrawObj>();
		for (int i=0; i < plListSize; ++i)
		{
			long lNewID = plChildList[i];
			DrawObj *poNewObj = new DrawObj(mpoPlane, lNewID);

			long lClass = lCheckTemplate(lNewID, TAG_clas, OC_DRAW);
			switch (lClass) {
			case OC_DRAW: 
				//draw objects
				poNewObj->vSetSelectable(FALSE);
				poNewObj->vSetLayer(DP_GROUND);
				break;
			case OC_FLOOR: 
				//'floor' objects
				poNewObj->vSetLayer(DP_FLOOR);
				break;
			case OC_OBSTACLE:
				//obstacles
				poNewObj->vSetLayer(DP_OBSTACLES);
				break;
			case OC_BUILDING:
				//buildings
				poNewObj->vSetLayer(DP_OBSTACLES); 
				break;
			case OC_ITEM:
				//items
				poNewObj->vSetLayer(DP_ITEMS);
				assert(0); //TBD - can items be children????
				break;
			}

			poNewObj->vSetParent(this);
			mpoChildren->AddToEnd(poNewObj);
		}
	}
}

void DrawObj::vSetParentOpen(Boolean bOpen)
{
	mbParentOpen = bOpen;

	if (bChildren() == FALSE)
		return;	//no kids to set so abort
	
	for (int i = 0; i < mpoChildren->GetSize(); ++i)
	{
		DrawObj *poObj = mpoChildren->GetItem(i);
		poObj->vSetParentOpen(bOpen);
	}

}

//Returns F if mpoChildren is NULL or if size of list is 0. 
//Returns T if at least 1 item in list
Boolean	DrawObj::bChildren()
{
	if (mpoChildren == NULL)
		return FALSE;

	if (mpoChildren->GetSize() >= 1)
		return TRUE;
	else
		return FALSE;
}

//Sets up nodes of children for the 1st time using node offsets in 'ctno' list
void DrawObj::vSetChildNodeNew(MapNode *poNode)
{

	long plListSize;
	long *plOffsetList;
	List<DrawObj> oTempDelete;	//a temporary list that will delete invalid objects once we're done
	plOffsetList = mpoTemplate->plGetList(TAG_ctno, &plListSize);
	if (plListSize) {
		assert ((plListSize % 2) == 0); //assert if we have an odd number of offsets. We should always have an even number
		for (int i=0; i < plListSize; i+=2)
		{
			int nOffX = plOffsetList[i];
			int nOffY = plOffsetList[i+1];
			//MapNode *poNewNode = mpoPlane->mpoMap->poGetNodeRC(poNode->mnXPos + nOffX, poNode->mnYPos + nOffY);
			MapNode *poNewNode = mpoPlane->mpoMap->poClosestNode(poNode->mnXPos + nOffX, poNode->mnYPos + nOffY);
			DrawObj *poChild = mpoChildren->GetItem(i/2);
			assert(poChild);
			assert(poNewNode);
//			if (poNewNode) 
//			{
				poChild->vSetNode(poNewNode);
				poNewNode->vAddObj(poChild);
//			} else {
//				if (gbMapEditing) 
//				{
//					//poChild->vSetNode(poChild->mpoParent->poNode());
//					//poChild->mpoParent->poNode()->vAddObj(poChild);
//				} else {
//					//probably got here by externally modifying the parents template value incorrectly
//					//
//					//starting position would be off the map somewhere so add to list
//					assert(0);	//lets assert just to tell us
//					oTempDelete.AddToEnd(mpoChildren->GetItem(i/2));
//					vDebugOut("Child had to be removed\n");
//				}
//			}

		}
	} else {
		//we're missing the Template Node Offset(ctno) list
		assert(0);
	}
	
	//oTempDelete will now delete anything in its list that may be been put there be an illegal template value
}

//Changes nodes of children(offset-wise) based on their offsets(ctno) from their parent
//in relation to poNewNode. 
//If after the offsets are calcuated and the child would end up off of the map somewhere,
//the child will wind up in the closest spot on the map edge that it can, but it won't go off the edge.
//
void DrawObj::vSetChildNodeOff(MapNode *poNewNode)
{

	for (int i = 0; i < mpoChildren->GetSize(); ++i)
	{
		DrawObj *poObj = mpoChildren->GetItem(i);
		assert(poObj);
		long plListSize;
		long *plOffsetList;
		plOffsetList = mpoTemplate->plGetList(TAG_ctno, &plListSize);
		if (plListSize) 
		{
			int nNodeOffX = plOffsetList[i*2];
			int nNodeOffY = plOffsetList[(i*2) + 1];
			MapNode *poNode = mpoPlane->mpoMap->poClosestNode(poNewNode->mnXPos + nNodeOffX, poNewNode->mnYPos + nNodeOffY);
			assert(poNode); //this should never assert
			poObj->vSetNode(poNode);
			poNode->vAddObj(poObj);
		}
	}
}

//Returns T if event was on child and that child bubbles up events, F for all other cases
Boolean DrawObj::bOnChildren(sciEvent &oEvent)
{
	if (mpoChildren)
	{
		for (int i = 0; i < mpoChildren->GetSize(); ++i)
		{
			DrawObj *poObj = mpoChildren->GetItem(i);
			
			//see if this child has any children that bubbles up events
			if (poObj->bOnChildren(oEvent))
				return TRUE;

			if (poObj->bOnMe(oEvent))
				if (poObj->bEventBubble())
					return TRUE;
				else
					return FALSE;
		}
	}
	return FALSE;
}

FLOAT DrawObj::vDamage(FLOAT fDamage, Unit * poUnitFrom)
{
	FLOAT fReturnDifference = mfHealth;

	mfHealth = max(0, mfHealth - fDamage);
	fReturnDifference = fReturnDifference - mfHealth;
	
	if (mfHealth == 0) 
	{
		if (bIsBuilding() || meClass == OC_DRAW)
		{
			// TBD 
			Missile * poMissile = new Missile(MISSILE_ANIMATEANDFADEDRAWOBJ, gpoMap->mpoPlane, (Unit*)this, mpoCurNode, mpoCurNode, 6021, 10, 0, 0, NULL, nPixelX(), nPixelY(), 1, 1);
			poMissile->SetDrawObj(nGetID());

			if (bIsBuilding())
			{
				// LARGER EXPLOSION
				poMissile = new Missile(MISSILE_ANIMATEANDFADEDRAWOBJ, gpoMap->mpoPlane, (Unit*)this, mpoCurNode, mpoCurNode, 6022, 10, 0, 0, NULL, nPixelX()-40, nPixelY()-20, 1, 1);
				poMissile = new Missile(MISSILE_ANIMATEANDFADEDRAWOBJ, gpoMap->mpoPlane, (Unit*)this, mpoCurNode, mpoCurNode, 6022, 10, 0, 0, NULL, nPixelX()+40, nPixelY()-30, 1, 1);
			}
		}
		else
		{
			DestroyFromMap();
		}
	}
	return fReturnDifference;
}

void DrawObj::DestroyFromMap()
{
	mpoCurNode->vClearCover();
	vClearObstacles(gpoMap, gpoMap->nGetWidth(), gpoMap->nGetHeight());
			
	if (bIsBuilding())
	{
		vSetBitmap(mpoTemplate->lGetValue(TAG_Bded, 67701));
		vSetLayer(DP_FLOOR);
		if (mbSelected) 
		{
			gpoMap->vRemoveSelection(this);
			gpoMap->vCommitSelection();
			vMarkUpdate();
		}
	}
	else
	{
		vSetBitmap(mpoTemplate->lGetValue(TAG_Bded, nBitmapID()));
		//vSetLayer(DP_GROUND);
		//mpoCurNode->vAddObjToFront(this);

		if (mbSelected) 
		{
			gpoMap->vRemoveSelection(this);
			gpoMap->vCommitSelection();
			vMarkUpdate();
		}

		mbVisible = FALSE;
	}
}

void DrawObj::vIncrementLoop()
{
	if (mnLoop + 1 >= nGetNumLoops())
		mnLoop = 0;
	else
		++mnLoop;

}

void DrawObj::vDecrementLoop()
{
	if (mnLoop - 1 < 0)
		mnLoop = nGetNumLoops() - 1;
	else
		--mnLoop;

}
