// FRAMEBUF.CPP -- The main frame buffer and interface to the screen
//
// DJM

#include "app.h"
#include "framebuf.h"
#include "window.h"
#include "util.h"
#include "rmerror.h"
#include "ztimer.h"

Blender *gpoBlender = NULL;
BYTE *spAlphaTable = NULL;

WORD *gpwRandPixels;

DDBLTFX sDDBLTFX;
DDSURFACEDESC sDDSD;

#undef TIME_BLITS

#define NUM_DRAW_TIMES 15
int		snCurDrawDelta = 0;
DWORD	sdwLastDrawTime = 0;
BOOL	sbCalcFrameRate = TRUE;
DWORD	sadwDrawDeltas[NUM_DRAW_TIMES];

FrameBuffer::FrameBuffer(int nWidth, int nHeight, int nBits, FB_TYPE eFBType)
{
	// for now, only allow 640x480x16
	assert(nWidth == 640);
	assert(nHeight == 480);
	assert(nBits == 16);

	// some static structures to avoid having to init them with every call
	ZeroMemory(&sDDBLTFX, sizeof(sDDBLTFX));
	sDDBLTFX.dwSize = sizeof(sDDBLTFX);
	ZeroMemory(&sDDSD, sizeof(sDDSD));
	sDDSD.dwSize = sizeof(sDDSD);

	ZeroMemory(&sadwDrawDeltas, NUM_DRAW_TIMES * sizeof(DWORD));

	mnWidth = nWidth;
	mnHeight = nHeight;
	mnBits = nBits;
	meFBType = eFBType;

	mpoPal16 = NULL;
	mhDIB = NULL;
	mpPixels = NULL;

	mlpDD = NULL;
	mlpDDSPrimary = NULL;
	mlpDDSBack = NULL;
	mlpDDSMemory = NULL;

	mpDIBPixels = NULL;			// the original origin of the dib pixels
	ZeroMemory(&mBMInfo, sizeof(mBMInfo));

	mbFlip = TRUE;

	switch (meFBType) {
	case FB_WINDOW:
		vInitWindow();
		break;

	case FB_FULLSCREEN:
		vInitDD();
	}

	// TEMP TEST --DJM

	gpwRandPixels = new WORD[500];
	int i;
	for (i = 0; i < 500; ++i) {
		gpwRandPixels[i] = pApp->oLocalRandom.nGetRandom(32768);
	}

}

FrameBuffer::~FrameBuffer()
{
	vDestroyWindowDIB();
	vCloseDD();

	if (mpoPal16) {
		delete mpoPal16;
	}
	if (gpoBlender) {
		delete gpoBlender;
	}
}

void FrameBuffer::vInitWindow(void)
{
	PF_MASKS Masks;

	int nScreenBits = nGetDisplayFormat(&Masks);
	if (nScreenBits != mnBits) {
		// throw a hissy fit
		MessageBox(NULL, "Vampire Wars must be run in 16-bit color mode.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		_exit(3);
	}
	if (!pApp->pWindow->hWnd) {
		pApp->pWindow->MakeWindow(mnWidth, mnHeight, WIN_FORCE_BORDERLESS);
	}

	if (gpoBlender) {
		delete gpoBlender;
	}
	gpoBlender = new Blender(Masks);
	vCreateCompatibleDIB();
}

void FrameBuffer::vDestroyWindowDIB(void)
{
	if (mhDIB) {
		DeleteObject(mhDIB);
		mhDIB = NULL;
		mpPixels = NULL;
		mpDIBPixels = NULL;
	}
}

HRESULT DirectDrawCreate2( GUID FAR *lpGUID, LPDIRECTDRAW FAR *lplpDD, IUnknown FAR *pUnkOuter  ) 
{
	HRESULT hResult;

	hResult = CoCreateInstance(CLSID_DirectDraw,
                                       NULL,
                                       CLSCTX_INPROC_SERVER,
                                       IID_IDirectDraw,
                                       (void**)lplpDD);

	if( !FAILED(hResult) )
	{
		hResult = (*lplpDD)->Initialize(NULL);	
	}

	return hResult;
}

void FrameBuffer::vInitDD(void)
{
	HRESULT ddrval;
	PF_MASKS Masks;

	
	if (!pApp->pWindow->hWnd) {
		pApp->pWindow->MakeWindow(mnWidth, mnHeight, WIN_FORCE_BORDERLESS);
	}

	// get a direct draw object
	ddrval = DirectDrawCreate2(NULL, &mlpDD, NULL);

	if (ddrval != DD_OK) {
		vDebugOut(D3DRMErrorToString(ddrval));
		// throw a hissy fit
		MessageBox(NULL, "Can't create DirectDraw object. Check to make sure you have the latest DD5.0 or higher drivers.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		exit(3);
	}

	// make us king
	ddrval = mlpDD->SetCooperativeLevel(pApp->pWindow->hWnd, DDSCL_EXCLUSIVE |
										DDSCL_FULLSCREEN);
	if (ddrval != DD_OK) {
		vDebugOut(D3DRMErrorToString(ddrval));
		MessageBox(NULL, "Can't take total control of screen. Check to make sure you have the latest DD5.0 or higher drivers, and that no other DirectDraw programs are running.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		exit(3);
	}

	ddrval = mlpDD->SetDisplayMode(mnWidth, mnHeight, mnBits);
	if (ddrval != DD_OK) {
		vDebugOut(D3DRMErrorToString(ddrval));
		char szErr[256];
		sprintf(szErr, "Can't set screen to %dx%d, %d bits. This is a common screen size and bit depth. Perhaps you need new drivers? Check the readme file for troubleshooting tips.", mnWidth, mnHeight, mnBits);
		MessageBox(NULL, szErr, "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		exit(3);
	}

	// Create the primary surface
	DDSURFACEDESC ddsd;

	if (mbFlip) {
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
			DDSCAPS_FLIP | DDSCAPS_COMPLEX;

		ddsd.dwBackBufferCount = 1;

		ddrval = mlpDD->CreateSurface(&ddsd, &mlpDDSPrimary, NULL);
		if (ddrval != DD_OK) {
			vDebugOut(D3DRMErrorToString(ddrval));
			MessageBox(NULL, "Can't create front and back buffer. Check to make sure you have the latest DD5.0 or higher drivers, and that no other DirectDraw programs are running.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
			exit(3);
		}
		// get pointer to back buffer
		ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
		ddrval = mlpDDSPrimary->GetAttachedSurface(&ddsd.ddsCaps, &mlpDDSBack);
		if (ddrval != DD_OK) {
			vDebugOut(D3DRMErrorToString(ddrval));
			MessageBox(NULL, "Can't get a pointer to back buffer. Check to make sure your system is not posessed by demons.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
			exit(3);
		}
	} else {
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

		ddrval = mlpDD->CreateSurface(&ddsd, &mlpDDSPrimary, NULL);
		if (ddrval != DD_OK) {
			vDebugOut(D3DRMErrorToString(ddrval));
			MessageBox(NULL, "Can't create primary buffer. Check to make sure you have the latest DD5.0 or higher drivers, and that no other DirectDraw programs are running.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
			exit(3);
		}
		mlpDDSBack = NULL;
	}

	// get the pixel format
	// DDPIXELFORMAT ddpf;
	ddrval = mlpDDSPrimary->GetSurfaceDesc(&ddsd);
	if (ddrval != DD_OK) {
		vDebugOut(D3DRMErrorToString(ddrval));
		MessageBox(NULL, "Can't get pixel format. Check to make sure your system is not posessed by demons.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		exit(3);
	}
	if (!(ddsd.ddpfPixelFormat.dwFlags & DDPF_RGB)) {
		MessageBox(NULL, "Pixel format not RGB. I'm afraid this game can only run on RGB display devices.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		exit(3);
	}
	Masks.dwRMask = ddsd.ddpfPixelFormat.dwRBitMask;
	Masks.dwGMask = ddsd.ddpfPixelFormat.dwGBitMask;
	Masks.dwBMask = ddsd.ddpfPixelFormat.dwBBitMask;

	if (gpoBlender) {
		delete gpoBlender;
	}
	gpoBlender = new Blender(Masks);

	mnPitch = ddsd.lPitch;

	// create the surface we'll be drawing to in memory
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
	ddsd.dwHeight = mnHeight;
	ddsd.dwWidth = mnWidth;
	ddrval = mlpDD->CreateSurface(&ddsd, &mlpDDSMemory, NULL);
	if (ddrval != DD_OK) {
		vDebugOut(D3DRMErrorToString(ddrval));
		MessageBox(NULL, "Can't create offscreen bitmap. Check to make sure your system is not posessed by demons.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		exit(3);
	}
	// keep the buffer locked all the time
	//mlpDDSMemory->Lock(NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT, NULL);
	//mnPitch = ddsd.lPitch;
	//mpPixels = ddsd.lpSurface;
}

void FrameBuffer::vCloseDD(void)
{
	if (!mlpDD) {
		return;
	}
	// unlock the buffer
	mlpDDSMemory->Unlock(NULL);
	mpPixels = NULL;
	mlpDD->RestoreDisplayMode();
	mlpDD->SetCooperativeLevel(pApp->pWindow->hWnd, DDSCL_NORMAL);

	//pApp->pWindow->vRestoreFromUL();

	if (mlpDD != NULL) {
		if (mlpDDSPrimary != NULL) {
			mlpDDSPrimary->Release();
			mlpDDSPrimary = NULL;
			mlpDDSBack = NULL;
		}
		if (mlpDDSMemory != NULL) {
			mlpDDSMemory->Release();
			mlpDDSMemory = NULL;
		}
		mlpDD->Release();
		mlpDD = NULL;
	}
}

void FrameBuffer::vSetMode(FB_TYPE eFBType)
{
	if (eFBType == meFBType) {
		return;
	}
	switch (meFBType) {
	case FB_WINDOW:
		vDestroyWindowDIB();
		break;

	case FB_FULLSCREEN:
		vCloseDD();
		break;
	}

	meFBType = eFBType;

	switch (meFBType) {
	case FB_WINDOW:
		vInitWindow();
		break;

	case FB_FULLSCREEN:
		vInitDD();
		break;
	}
#ifdef TIME_BLITS
	TIMER_RESET();
#endif
}

void FrameBuffer::vCreateCompatibleDIB(void)
{
	ZeroMemory(&mBMInfo, sizeof(mBMInfo));
	mBMInfo.bi.biSize = sizeof(mBMInfo.bi);
	mBMInfo.bi.biBitCount = 0;

	HDC hdcDisplay = GetDC(NULL);
	HBITMAP hbm = CreateCompatibleBitmap(hdcDisplay, 1, 1);
	GetDIBits(hdcDisplay, hbm, 0, 1, NULL, (BITMAPINFO *) & mBMInfo, DIB_RGB_COLORS);

	// we need to call twice to get the masks. Why? Because someone at MS
	// hates us.
	GetDIBits(hdcDisplay, hbm, 0, 1, NULL, (BITMAPINFO *) & mBMInfo, DIB_RGB_COLORS);
	DeleteObject(hbm);
	ReleaseDC(NULL, hdcDisplay);

	mBMInfo.bi.biWidth = mnWidth;
	mBMInfo.bi.biHeight = mnHeight;

	mhDIB = CreateDIBSection(pApp->pWindow->hDC, (BITMAPINFO *) & mBMInfo,
							 DIB_RGB_COLORS, &mpDIBPixels, NULL, 0);
	if (!mhDIB) {
		MessageBox(NULL, "Unable to create working DIB.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		assert(0);
		return;
	}
	// since this is a bottom-up DIB, instead use the exciting DirectDraw
	// method
	// of pointing to the upper-left corner and making the pitch negative
	switch (mnBits) {
	case 8:
		mnPitch = -mnWidth;
		mpPixels = (void *) (((BYTE *) mpDIBPixels) + ((mnHeight - 1) * mnWidth));
		break;
	case 16:
		mnPitch = -(mnWidth * 2);
		mpPixels = (void *) (((WORD *) mpDIBPixels) + ((mnHeight - 1) * mnWidth));
		break;
	case 32:
		mnPitch = -(mnWidth * 4);
		mpPixels = (void *) (((DWORD *) mpDIBPixels) + ((mnHeight - 1) * mnWidth));
		break;
	default:
		MessageBox(NULL, "FrameBuffer only works with 8. 16, and 32 bit dibs.", "Error", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
		assert(0);
		return;
		break;
	}
}


void FrameBuffer::vSetPalette(Palette16 * poPal)
{
	if (mpoPal16) {
		delete mpoPal16;
	}
	mpoPal16 = poPal;
}

//void FrameBuffer::vGetFrameBufferInfo(BUFFER_INFO *pInfo);

/*
void *FrameBuffer::pGetBuffer(void) // get a writeable pointer to the frame buffer
{
	switch (meFBType) {
	case FB_WINDOW:
		return mpPixels;
		break;

	case FB_FULLSCREEN:
		return mpPixels;
		break;
	}
	return NULL;
}

void FrameBuffer::vReleaseBuffer(void) // done writing to the buffer
{
	// do nothing
}
*/

void FrameBuffer::vLockBuffer(void)
{
	HRESULT ddrval;
	if (meFBType == FB_FULLSCREEN) {
		// re-lock the buffer
		// keep the buffer locked all the time
		ddrval = mlpDDSMemory->Lock(NULL, &sDDSD, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT, NULL);
#ifdef _DEBUG
		if (ddrval) {
			vDebugOut(D3DRMErrorToString(ddrval));
			assert(0);
		}
#endif
		mnPitch = sDDSD.lPitch;
		mpPixels = sDDSD.lpSurface;
	}
}

void FrameBuffer::vFrameDone(void)
{								// write the frame out to the screen
	HDC hDC;
	HRESULT ddrval;

#ifdef TIME_BLITS
	TIMER_BEGIN();
#endif

	switch (meFBType) {
	case FB_WINDOW:
		assert(pApp && pApp->pWindow);
		hDC = pApp->pWindow->hGetDC();
		SetDIBitsToDevice(hDC,
						  0, 0,	// dest pos
						  mnWidth, mnHeight,	// size
						  0, 0,	// source pos
						  0, mnHeight,	// starting scanline, number of
										// scanlines
				  mpDIBPixels, (BITMAPINFO *) & mBMInfo, DIB_RGB_COLORS);
		pApp->pWindow->vReleaseDC();
		break;

	case FB_FULLSCREEN:
		// unlock the buffer
		ddrval = mlpDDSMemory->Unlock(NULL);
#ifdef _DEBUG
		if (ddrval) {
			vDebugOut(D3DRMErrorToString(ddrval));
			assert(0);
		}
#endif
		mpPixels = NULL;

		if (mbFlip) {
			// blit from memory to back buffer
			ddrval = mlpDDSBack->Blt(NULL, mlpDDSMemory, NULL, DDBLT_WAIT, &sDDBLTFX);
#ifdef _DEBUG
			if (ddrval) {
				vDebugOut(D3DRMErrorToString(ddrval));
			}
#endif

			// flip the pages
			ddrval = mlpDDSPrimary->Flip(NULL, NULL);
#ifdef _DEBUG
			if (ddrval) {
				vDebugOut(D3DRMErrorToString(ddrval));
			}
#endif
		} else {
			// blit from memory to front buffer
			ddrval = mlpDDSPrimary->Blt(NULL, mlpDDSMemory, NULL, NULL, &sDDBLTFX);
#ifdef _DEBUG
			if (ddrval) {
				vDebugOut(D3DRMErrorToString(ddrval));
			}
#endif
			if (ddrval == DDERR_SURFACELOST) {
				ddrval = mlpDDSPrimary->Restore();
#ifdef _DEBUG
				if (ddrval) {
					vDebugOut(D3DRMErrorToString(ddrval));
				}
#endif
				ddrval = mlpDDSPrimary->Blt(NULL, mlpDDSMemory, NULL, NULL, &sDDBLTFX);
#ifdef _DEBUG
				if (ddrval) {
					vDebugOut(D3DRMErrorToString(ddrval));
				}
#endif
			}

		}

		break;
	}

#ifdef TIME_BLITS
	TIMER_END();
#endif

	if (sbCalcFrameRate) {
		DWORD dwCurDrawTime = GetTickCount();
		DWORD dwDiff = dwCurDrawTime - sdwLastDrawTime;
		sdwLastDrawTime = dwCurDrawTime;

		// put it into the deltas list
		sadwDrawDeltas[snCurDrawDelta] = dwDiff;
		snCurDrawDelta++;
		if (snCurDrawDelta == NUM_DRAW_TIMES) {
			snCurDrawDelta = 0;
		}

		// now average all deltas
		int nSize = NUM_DRAW_TIMES;
		DWORD dwAccum = 0;
		while (nSize--) {
			dwAccum += sadwDrawDeltas[nSize];
		}
		dwAccum /= NUM_DRAW_TIMES;

		// and calculate a final times/second
		pApp->mfFPS	= 1000.0 / dwAccum;
	}
}

/********************************************************
 *
 *  BLENDER
 *
 ********************************************************/
COLOR16 FASTCALL dwPack565(BYTE jRed, BYTE jGreen, BYTE jBlue)
{
	return 
		((((COLOR16)jRed)   << 8) & 0xF800L) |
		((((COLOR16)jGreen) << 3) & 0x07E0L) |
		((((COLOR16)jBlue)  >> 3));

}

COLOR16 FASTCALL dwPack555(BYTE jRed, BYTE jGreen, BYTE jBlue)
{
	return 
		((((COLOR16)jRed)   << 7) & 0x7C00L) |
		((((COLOR16)jGreen) << 2) & 0x03E0L) |
		((((COLOR16)jBlue)  >> 3));

}

COLOR16 FASTCALL dwPackGeneric(BYTE jRed, BYTE jGreen, BYTE jBlue)
{
	return 
		((((COLOR16)jRed)   >> gpoBlender->mjRBaseShift) << gpoBlender->mjRShift) |
		((((COLOR16)jGreen) >> gpoBlender->mjGBaseShift) << gpoBlender->mjGShift) |
		((((COLOR16)jBlue)  >> gpoBlender->mjBBaseShift) << gpoBlender->mjBShift);

}

//
// CalcAlphaBlack
//
// Calculates Alpha assuming that the background is black, which allows
// some mighty big speed wins
//

COLOR16 FASTCALL dwCalcAlphaBlackGeneric(COLOR16 dwFore, ALPHA5 jAlpha) {

	// we don't do anything special for full and none, as we probably
	// won't have any pixels with those values -- they just won't
	// involve alpha and this function won't be called.

	DWORD dwRet = spAlphaTable[((_rotl(dwFore, gpoBlender->mnRRotToFore) & AT_FORE_MASK) |
				  jAlpha)] << gpoBlender->mjRShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnGRotToFore) & AT_FORE_MASK) |
				  jAlpha)] << gpoBlender->mjGShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnBRotToFore) & AT_FORE_MASK) |
				  jAlpha)] << gpoBlender->mjBShift5to16;
	
	return dwRet;
}

COLOR16 FASTCALL dwCalcAlphaBlack565(COLOR16 dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 565 pixel format

	return	(spAlphaTable[((dwFore >> 1) & AT_FORE_MASK) |
			jAlpha] << 11) |
			(spAlphaTable[((dwFore << 4) & AT_FORE_MASK) |
			jAlpha] << 6) |
			(spAlphaTable[((dwFore << 10) & AT_FORE_MASK) |
			jAlpha]);
	
}

COLOR16 FASTCALL dwCalcAlphaBlack555(COLOR16 dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 555 pixel format

	return	(spAlphaTable[(dwFore & AT_FORE_MASK) |
			jAlpha] << 10) |
			(spAlphaTable[((dwFore << 5) & AT_FORE_MASK) |
			jAlpha] << 5) |
			(spAlphaTable[((dwFore << 10) & AT_FORE_MASK) |
			jAlpha]);
	
}


//
// Calc2AlphaBlack
//
// calculates 2 pixels at a time for some extra bonus wins
//
DWORD FASTCALL dwCalc2AlphaBlackGeneric(DWORD dwFore, ALPHA5 jAlpha) {

	// we don't do anything special for full and none, as we probably
	// won't have any pixels with those values -- they just won't
	// involve alpha and this function won't be called.

	DWORD dwRet = spAlphaTable[((_rotl(dwFore, gpoBlender->mnRRotToFore) & AT_FORE_MASK) |
				  jAlpha)] << gpoBlender->mjRShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnGRotToFore) & AT_FORE_MASK) |
				  jAlpha)] << gpoBlender->mjGShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnBRotToFore) & AT_FORE_MASK) |
				  jAlpha)] << gpoBlender->mjBShift5to16;
	dwRet	   |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnRRotToFore + 16) & AT_FORE_MASK) |
				  jAlpha)] << (gpoBlender->mjRShift5to16 + 16);
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnGRotToFore + 16) & AT_FORE_MASK) |
				  jAlpha)] << (gpoBlender->mjGShift5to16 + 16);
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnBRotToFore + 16) & AT_FORE_MASK) |
				  jAlpha)] << (gpoBlender->mjBShift5to16 + 16);
	
	return dwRet;
}

DWORD FASTCALL dwCalc2AlphaBlack565(DWORD dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 565 pixel format

	return	(spAlphaTable[((dwFore >> 1) & AT_FORE_MASK) |
			jAlpha] << 11) |
			(spAlphaTable[((dwFore << 4) & AT_FORE_MASK) |
			jAlpha] << 6) |
			(spAlphaTable[((dwFore << 10) & AT_FORE_MASK) |
			jAlpha]) |
			(spAlphaTable[((dwFore >> 17) & AT_FORE_MASK) |
			jAlpha] << 27) |
			(spAlphaTable[((dwFore >> 12) & AT_FORE_MASK) |
			jAlpha] << 22) |
			(spAlphaTable[((dwFore >> 6) & AT_FORE_MASK) |
			jAlpha] << 16);
	
}

DWORD FASTCALL dwCalc2AlphaBlack555(DWORD dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 555 pixel format

	return	(spAlphaTable[(dwFore & AT_FORE_MASK) |
			jAlpha] << 10) |
			(spAlphaTable[((dwFore << 5) & AT_FORE_MASK) |
			jAlpha] << 5) |
			(spAlphaTable[((dwFore << 10) & AT_FORE_MASK) |
			jAlpha]) |
			(spAlphaTable[((dwFore >> 16) & AT_FORE_MASK) |
			jAlpha] << 26) |
			(spAlphaTable[((dwFore >> 11) & AT_FORE_MASK) |
			jAlpha] << 21) |
			(spAlphaTable[((dwFore >> 6) & AT_FORE_MASK) |
			jAlpha] << 16); 
	
}

//
// CalcAlpha
//
// does an alpha calc through a lookup table using 5-bit accuracy

COLOR16 FASTCALL dwCalcAlphaGeneric(COLOR16 dwBack, COLOR16 dwFore, ALPHA5 jAlpha) {

	// we don't do anything special for full and none, as we probably
	// won't have any pixels with those values -- they just won't
	// involve alpha and this function won't be called.

	DWORD dwRet = spAlphaTable[((_rotl(dwFore, gpoBlender->mnRRotToFore) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnRRotToBack) & AT_BACK_MASK) |
				  jAlpha)] << gpoBlender->mjRShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnGRotToFore) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnGRotToBack) & AT_BACK_MASK) |
				  jAlpha)] << gpoBlender->mjGShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnBRotToFore) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnBRotToBack) & AT_BACK_MASK) |
				  jAlpha)] << gpoBlender->mjBShift5to16;
	
	return dwRet;
}

COLOR16 FASTCALL dwCalcAlpha565(COLOR16 dwBack, COLOR16 dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 565 pixel format

	return	(spAlphaTable[((dwFore >> 1) & AT_FORE_MASK) |
			((dwBack >> 6) & AT_BACK_MASK) |
			jAlpha] << 11) |
			(spAlphaTable[((dwFore << 4) & AT_FORE_MASK) |
			((dwBack >> 1) & AT_BACK_MASK) |
			jAlpha] << 6) |
			(spAlphaTable[((dwFore << 10) & AT_FORE_MASK) |
			((dwBack << 5) & AT_BACK_MASK) |
			jAlpha]);
	
}

COLOR16 FASTCALL dwCalcAlpha555(COLOR16 dwBack, COLOR16 dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 555 pixel format

	// convert alpha to alpha5
	return	(spAlphaTable[(dwFore & AT_FORE_MASK) |
			((dwBack >> 5) & AT_BACK_MASK) |
			jAlpha] << 10) |
			(spAlphaTable[((dwFore << 5) & AT_FORE_MASK) |
			(dwBack & AT_BACK_MASK) |
			jAlpha] << 5) |
			(spAlphaTable[((dwFore << 10) & AT_FORE_MASK) |
			((dwBack << 5) & AT_BACK_MASK) |
			jAlpha]);
	
}

//
// Calc2Alpha
//
// Calculates 2 pixels at a time, for some nice wins
DWORD FASTCALL dwCalc2AlphaGeneric(DWORD dwBack, DWORD dwFore, ALPHA5 jAlpha) {

	// we don't do anything special for full and none, as we probably
	// won't have any pixels with those values -- they just won't
	// involve alpha and this function won't be called.

	DWORD dwRet = spAlphaTable[((_rotl(dwFore, gpoBlender->mnRRotToFore) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnRRotToBack) & AT_BACK_MASK) |
				  jAlpha)] << gpoBlender->mjRShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnGRotToFore) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnGRotToBack) & AT_BACK_MASK) |
				  jAlpha)] << gpoBlender->mjGShift5to16;
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnBRotToFore) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnBRotToBack) & AT_BACK_MASK) |
				  jAlpha)] << gpoBlender->mjBShift5to16;

	dwRet	   |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnRRotToFore + 16) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnRRotToBack + 16) & AT_BACK_MASK) |
				  jAlpha)] << (gpoBlender->mjRShift5to16 + 16);
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnGRotToFore + 16) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnGRotToBack + 16) & AT_BACK_MASK) |
				  jAlpha)] << (gpoBlender->mjGShift5to16 + 16);
	dwRet      |= spAlphaTable[((_rotl(dwFore, gpoBlender->mnBRotToFore + 16) & AT_FORE_MASK) |
				  (_rotl(dwBack, gpoBlender->mnBRotToBack + 16) & AT_BACK_MASK) |
				  jAlpha)] << (gpoBlender->mjBShift5to16 + 16);
	
	return dwRet;
}

DWORD FASTCALL dwCalc2Alpha565(DWORD dwBack, DWORD dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 555 pixel format

	// convert alpha to double-alpha
	DWORD dwAlpha = jAlpha | ((DWORD)jAlpha << 16);

	DWORD dwTemp = ((dwFore >> 1) & AT_2_FORE_MASK) |
			((dwBack >> 6) & AT_2_BACK_MASK) |
			dwAlpha;
	DWORD dwRet = (spAlphaTable[dwTemp >> 16] << 27) |
		(spAlphaTable[dwTemp & 0xFFFF] << 11);

	dwTemp = ((dwFore << 4) & AT_2_FORE_MASK) |
			((dwBack >> 1) & AT_2_BACK_MASK) |
			dwAlpha;
	dwRet |= (spAlphaTable[dwTemp >> 16] << 22) |
		(spAlphaTable[dwTemp & 0xFFFF] << 6);

	dwTemp = ((dwFore << 10) & AT_2_FORE_MASK) |
			((dwBack << 5) & AT_2_BACK_MASK) |
			dwAlpha;
	dwRet |= (spAlphaTable[dwTemp >> 16] << 16) |
		(spAlphaTable[dwTemp & 0xFFFF]);

	return	dwRet;
	
}

DWORD FASTCALL dwCalc2Alpha555(DWORD dwBack, DWORD dwFore, ALPHA5 jAlpha) {

	// this is hand-tuned for 555 pixel format

	// convert alpha to double-alpha
	DWORD dwAlpha = jAlpha | ((DWORD)jAlpha << 16);

	DWORD dwTemp = (dwFore & AT_2_FORE_MASK) |
			((dwBack >> 5) & AT_2_BACK_MASK) |
			dwAlpha;
	DWORD dwRet = (spAlphaTable[dwTemp >> 16] << 26) |
		(spAlphaTable[dwTemp & 0xFFFF] << 10);

	dwTemp = ((dwFore << 5) & AT_2_FORE_MASK) |
			(dwBack & AT_2_BACK_MASK) |
			dwAlpha;
	dwRet |= (spAlphaTable[dwTemp >> 16] << 21) |
		(spAlphaTable[dwTemp & 0xFFFF] << 5);

	dwTemp = ((dwFore << 10) & AT_2_FORE_MASK) |
			((dwBack << 5) & AT_2_BACK_MASK) |
			dwAlpha;
	dwRet |= (spAlphaTable[dwTemp >> 16] << 16) |
		(spAlphaTable[dwTemp & 0xFFFF]);

	return	dwRet;
	
}


Blender::Blender(PF_MASKS & Masks)
{
	// get bit positions and lowest bit
	int nLowBit;
	int nBits;

	mdwBlack50Mask = 0;
	mdwBlack25Mask = 0;

	nBits = nMaskToNumBits(Masks.dwRMask, &nLowBit);
	mjRBaseShift = 8 - nBits;
	mjRShift = nLowBit;
	mjRShift5to16 = nLowBit + (nBits - 5);
	mnRRotToFore = (AT_FORE_LOWBIT - mjRShift5to16 + 32) % 32;
	mnRRotToBack = (AT_BACK_LOWBIT - mjRShift5to16 + 32) % 32;

	mdwBlack50Mask = 15 << mjRShift5to16;
	mdwBlack25Mask = 7  << mjRShift5to16;

	nBits = nMaskToNumBits(Masks.dwGMask, &nLowBit);
	mjGBaseShift = 8 - nBits;
	mjGShift = nLowBit;
	mjGShift5to16 = nLowBit + (nBits - 5);
	mnGRotToFore = (AT_FORE_LOWBIT - mjGShift5to16 + 32) % 32;
	mnGRotToBack = (AT_BACK_LOWBIT - mjGShift5to16 + 32) % 32;

	mdwBlack50Mask |= 15 << mjGShift5to16;
	mdwBlack25Mask |= 7  << mjGShift5to16;

	nBits = nMaskToNumBits(Masks.dwBMask, &nLowBit);
	mjBBaseShift = 8 - nBits;
	mjBShift = nLowBit;
	mjBShift5to16 = nLowBit + (nBits - 5);
	mnBRotToFore = (AT_FORE_LOWBIT - mjBShift5to16 + 32) % 32;
	mnBRotToBack = (AT_BACK_LOWBIT - mjBShift5to16 + 32) % 32;

	mdwBlack50Mask |= 15 << mjBShift5to16;
	mdwBlack25Mask |= 7  << mjBShift5to16;

	// copy the fast masks twice to make them capable of doing
	// two pixels at once
	mdwBlack50Mask |= mdwBlack50Mask << 16;
	mdwBlack25Mask |= mdwBlack25Mask << 16;

	vInitAlphaTable();

	spAlphaTable = majAlphaTable;

	if (Masks.dwRMask == 0xF800L &&
		Masks.dwGMask == 0x07E0L &&
		Masks.dwBMask == 0x001FL) {

		// we're in 565 RGB mode
		mpdwAlphaCalc	= dwCalcAlpha565;
		mpdw2AlphaCalc	= dwCalc2Alpha565;
		mpdwAlphaBlack	= dwCalcAlphaBlack565;
		mpdw2AlphaBlack	= dwCalc2AlphaBlack565;
		mpdwPack		= dwPack565;
	} else if (Masks.dwRMask == 0x7C00L &&
		Masks.dwGMask == 0x03E0L &&
		Masks.dwBMask == 0x001FL) {

		// we're in 555 RGB mode
		mpdwAlphaCalc	= dwCalcAlpha555;
		mpdw2AlphaCalc	= dwCalc2Alpha555;
		mpdwAlphaBlack	= dwCalcAlphaBlack555;
		mpdw2AlphaBlack	= dwCalc2AlphaBlack555;
		mpdwPack		= dwPack555;
	} else {

		// we're in some other mode, so use the slow version
		mpdwAlphaCalc	= dwCalcAlphaGeneric;
		mpdw2AlphaCalc	= dwCalc2AlphaGeneric;
		mpdwAlphaBlack	= dwCalcAlphaBlackGeneric;
		mpdw2AlphaBlack	= dwCalc2AlphaBlackGeneric;
		mpdwPack		= dwPackGeneric;
	}
}

Blender::~Blender()
{
}

void Blender::vInitAlphaTable(void)
{
	int i,
	 j,
	 k;
	for (i = 0; i < AM_COLOR_MAX; ++i) {
		for (j = 0; j < AM_COLOR_MAX; ++j) {
			for (k = 0; k < AM_ALPHA_MAX; ++k) {
				// we'll do the alpha calculation with lots of excess
				// precision
				// just to be sure -- time is not of the essence here
				double fFore = i;
				double fBack = j;
				double fAlpha = k / AM_ALPHA_DIVISOR;
				double fResult = (fFore * fAlpha) + (fBack * (1.0 - fAlpha));

				// round and truncate
				signed char jColor = (signed char) (fResult + 0.5);
				if (jColor >= AM_COLOR_MAX) {
					jColor = AM_COLOR_MAX - 1;
				} else if (jColor < 0) {
					jColor = 0;
				}
				vSetAlphaTable(i, j, k, jColor);
			}
		}
	}
}

//
// This function gets the display format, and returns the real bits per pixel.
//
// If the bits per pixel is 16 or above, it also fills in the mask parameters
// with the mask values for each color.
//
int nGetDisplayFormat(PF_MASKS * pMasks)
{
	MY_BM_INFO mBMInfo;

	ZeroMemory(&mBMInfo, sizeof(mBMInfo));
	mBMInfo.bi.biSize = sizeof(mBMInfo.bi);
	mBMInfo.bi.biBitCount = 0;

	HDC hdcDisplay = GetDC(NULL);
	HBITMAP hbm = CreateCompatibleBitmap(hdcDisplay, 1, 1);
	GetDIBits(hdcDisplay, hbm, 0, 1, NULL, (BITMAPINFO *) & mBMInfo, DIB_RGB_COLORS);

	// we need to call twice to get the masks. Why? Because someone at MS
	// hates us.
	GetDIBits(hdcDisplay, hbm, 0, 1, NULL, (BITMAPINFO *) & mBMInfo, DIB_RGB_COLORS);
	DeleteObject(hbm);
	ReleaseDC(NULL, hdcDisplay);

	if (mBMInfo.bi.biCompression != BI_BITFIELDS) {
		// Either BI_RGB or BI_RLE_something
		// .... or perhaps (!!) something else.
		// Theoretically biCompression might be
		// mmioFourCC('c','v','i','d') but I doubt it.
		if (mBMInfo.bi.biBitCount == 16 && mBMInfo.bi.biCompression == BI_RGB) {

			// the standard 555 format
			if (pMasks) {
				pMasks->dwRMask = 0x07C00;
				pMasks->dwGMask = 0x003E0;
				pMasks->dwBMask = 0x0001F;
			}
			return 16;
		} else {

			// 1, 2, 4, or 8 bit format
			if (pMasks) {
				pMasks->dwRMask = 0;
				pMasks->dwGMask = 0;
				pMasks->dwBMask = 0;
			}
			return mBMInfo.bi.biBitCount;
		}
	} else {
		if (pMasks) {
			pMasks->dwRMask = mBMInfo.fields[0];
			pMasks->dwGMask = mBMInfo.fields[1];
			pMasks->dwBMask = mBMInfo.fields[2];
		}
		DWORD allmasks = mBMInfo.fields[0] | mBMInfo.fields[1] | mBMInfo.fields[2];
		int k = 0;
		while (allmasks) {
			if (allmasks & 1)
				k++;
			allmasks /= 2;
		}

		// kludge to make it return 16 bits when it's really 15
		if (k == 15) {
			k = 16;
		}
		return k;
	}
}

int nMaskToNumBits(DWORD dwMask, int *pnLowBit)
{
	assert(dwMask);
	assert(pnLowBit);

	int nBits = 0;
	int nLowBit = 0;
	BOOL bFoundLowBit = FALSE;

	while (dwMask) {
		if (dwMask & 1) {
			if (!bFoundLowBit) {
				bFoundLowBit = TRUE;
			}
			nBits++;
		};

		dwMask >>= 1;
		if (!bFoundLowBit) {
			nLowBit++;
		}
	}

	*pnLowBit = nLowBit;

	return nBits;
}
