#include "os.h"
#include "list.h"
#include <string.h>

typedef Object *POBJECT;

/* Make a new list */
ListBase::ListBase()
{
	iListCount = 0;
	//DEFAULT_DELTA = DEFAULT_DELTA;
	//iNextIndex = 0;
	//iPreviousIndex = -1;

	/* Make space for a group of pointers */
	iSpaceAllocated = DEFAULT_DELTA;
	ppoList = new POBJECT[iSpaceAllocated];
	memset((BYTE*)ppoList, 0, iSpaceAllocated * sizeof(Object *));
}

ListBase::~ListBase()
{
	DisposeAll();
	if (ppoList) {
		delete[] ppoList;
	}
}


void ListBase::Grow(int nExtraNeeded)
{
	int oldSize,
	 newSize;
	Object **ppoNew;

	if (nExtraNeeded <= 0) {
		nExtraNeeded = DEFAULT_DELTA;
	}
	oldSize = iSpaceAllocated * sizeof(Object *);
	newSize = (iSpaceAllocated + nExtraNeeded) * sizeof(Object *);
	iSpaceAllocated = iSpaceAllocated + nExtraNeeded;
	ppoNew = new POBJECT[iSpaceAllocated];
	memset(ppoNew, 0, newSize);
	memcpy((BYTE*)ppoNew, (BYTE*)ppoList, oldSize);
	delete[] ppoList;
	ppoList = ppoNew;
}


void ListBase::Shrink(void)
{
	int oldSize,
	 newSize;
	Object **ppoNew;

	if (iSpaceAllocated >= iListCount + DEFAULT_DELTA) {
		return;
	}
	oldSize = iListCount * sizeof(Object *);

	// we'll go ahead and shrink it to a little larger than needed
	// so we don't immediately reallocate next time
	newSize = (iListCount + DEFAULT_DELTA) * sizeof(Object *);
	iSpaceAllocated = iListCount + DEFAULT_DELTA;
	ppoNew = new POBJECT[iSpaceAllocated];
	memset(ppoNew, 0, newSize);
	memcpy(ppoNew, ppoList, oldSize);
	delete[] ppoList;
	ppoList = ppoNew;
}

void ListBase::Preallocate(int nItems)
{
	if (iSpaceAllocated >= nItems) {
		return;
	}
	Grow(nItems - iSpaceAllocated);
}


void ListBase::AddToEnd(Object * pObject)
{
	if (iListCount == iSpaceAllocated) {	// Need more room
		Grow();
	}
	ppoList[iListCount++] = (Object *) pObject;
}


void ListBase::AddToFront(Object * pObject)
{
	Insert(pObject, 0);
}


void ListBase::AddAfter(Object * seekObject, Object * pObject)
{
	int nLoc;

	// Try to find the seekObject.
	nLoc = Find(seekObject);

	if (nLoc != ID_NONE) {
		// seekObject found
		Insert(pObject, nLoc + 1);
	}
}


int  ListBase::Find(Object *poThis)
{
	int i;

	for (i = 0; i < iListCount; ++ i) {
		if (poThis == ppoList[i]) {
			return i;
		}
	}

	return ID_NONE;
}


void ListBase::Insert(Object * pObject, int iIndex)
{
	int i;

	if ((iListCount == iSpaceAllocated) || (iIndex >= iSpaceAllocated)) {
		Grow(max(0, (iIndex - iSpaceAllocated) + DEFAULT_DELTA));
	}
	// if we're inserting off the end, just add the amount of space
	// necessary
	if (iIndex >= iListCount) {
		iListCount = iIndex + 1;
	} else {
		++iListCount;
	}

	// Open up a space
	//for (i = iListCount - 1; i >= iIndex; --i) {	  // PB - this was off by one
	for ( i = iListCount-1; i > iIndex; i--)
	{
		ppoList[i] = ppoList[i-1];
	}
	// ++iListCount;
	ppoList[iIndex] = (Object *) pObject;	// Put the pointer in the space
}

// NOTE: Remove and RemoveIndex do not dispose of the object.
//       To dispose the object, use Delete (TBD)

void ListBase::Remove(Object * pObject)
{
	Object *p;
	int i;

	for (i = 0; i < iListCount; ++i) {
		p = GetItem(i);
		if (p == pObject) {
			RemoveIndex(i);
			break;
		}
	}
}


void ListBase::RemoveIndex(int iIndex)
{
	int i;

	if (iListCount && (iIndex >= 0) && (iIndex < iListCount)) {
		--iListCount;
		for (i = iIndex; i < iListCount; ++i) {
			ppoList[i] = ppoList[i + 1];
		}

		// adjust next and previous, if necessary
		/*
		if (iNextIndex > iIndex) {
			--iNextIndex;
		}
		if (iPreviousIndex >= iIndex) {
			--iPreviousIndex;
		}
		*/

		// zero the last (invalid) entry
		ppoList[iListCount] = NULL;
	}
}


void ListBase::RemoveAll()
{
	// remove without disposing items
	// essentially re-init the list

	iListCount = 0;
	//iNextIndex = 0;
	//iPreviousIndex = -1;

	memset(ppoList, 0, iSpaceAllocated * sizeof(Object *));
}


Object *ListBase::GetFirst()
{
	//iNextIndex = 1;
	//iPreviousIndex = -1;
	if (iListCount) {
		return ppoList[0];
	} else {
		return NULL;
	}
}


Object *ListBase::GetLast()
{
	if (iListCount) {
		//iPreviousIndex = iListCount - 2;
		//iNextIndex = iListCount;
		return ppoList[iListCount - 1];
	} else {
		return NULL;
	}
}


// pops off the last item, removing it from the list.
// the opposite of AddToEnd.
Object *ListBase::PopLast()
{
	Object *po;
	if (iListCount) {
		po = GetLast();
		RemoveIndex(iListCount - 1);
	} else {
		po = NULL;
	}
	return po;
}

/*
Object *ListBase::GetNext()
{
	if (iNextIndex < iListCount) {
		Object *po = ppoList[iNextIndex];
		++iNextIndex;
		iPreviousIndex = iNextIndex - 2; // ensures re-sync after delete
		return po;
	} else {
		return NULL;
	}
}


Object *ListBase::GetPrevious()
{
	if (iPreviousIndex >= 0) {
		Object *po = ppoList[iPreviousIndex];
		--iPreviousIndex;
		iNextIndex = iPreviousIndex + 2; // ensures re-sync after delete
		return po;
	} else {
		return NULL;
	}
}
*/


void ListBase::DisposeAll()
{
	while (iListCount) {
		--iListCount;
		DisposeItem(ppoList[iListCount]);
		ppoList[iListCount] = NULL;
	}
}


void ListBase::DisposeItem(Object *po)
{
	if (po) {
		delete po;
	}
}

#if 0

Object *ListBase::GetItem(int iIndex)
{
	if (iIndex < iListCount) {
		return ppoList[iIndex];
	} else {
		return NULL;
	}
}
#endif


void ListBase::SetItem(int iIndex, Object * pObject)
{
	if (iIndex >= iSpaceAllocated) {
		Grow((iIndex - iSpaceAllocated) + DEFAULT_DELTA);
	}

	//DisposeItem(ppoList[iIndex]);

	ppoList[iIndex] = (Object *) pObject;
	if (iIndex >= iListCount) {
		iListCount = iIndex + 1;
	}
}



void ListBase::MoveToFront(Object * pObject)
{
	Remove(pObject);
	Insert(pObject, 0);
}


