//
// RANDGEN.H
//
// Two, count them, two random number generators from
// various books. Completely eliminates dependency on system-supplied
// random-number generators.
// 
// The r250 generator is very fast and very random, and depends on
// the linear congruent generator for its initialization
// 
// The RandomGenerator class uses r250 internally. It's the recommended one.
//
// Don Munsil. 4/10/97

#ifndef _RANDGEN_
#define _RANDGEN_

#include <stdio.h>

class RandomGenerator {
private:
	unsigned int	maunTable[250];
	int				mnIndex;

public:
	//
	// NOTE: initializing or reseeding the RandomGenerator
	// class re-seeds the linear congruent generator.
	// Generally speaking, there's no reason to use the LC
	// generator directly.
	//
	RandomGenerator(unsigned int unSeed = 1);

	//
	// for re-seeding the generator after init
	// note that this has the side effect of re-seeding
	// the LC generator in this module
	//
	void vSeed(unsigned int unSeed = 1);

	//
	// network debugging stuff -- checksum and text dumping
	//
	long lGetChecksum(void);
	void vDumpSyncText(FILE *fileDump);

	//
	// Wrapper functions -- useful functions that do cool things, though
	// all eventually call unGetRandom to do the dirty work
	//
	int nGetRandom(int nValues); // 0 to (nValues - 1)
	int nGetRandom(int nLower, int nUpper); // nLower to nUpper, inclusive
	double dGetRandom(void); // between 0.0 and 1.0, inclusive

	//
	// This is the main function. It's inlined for speed.
	//
	unsigned int unGetRandom(void) // fully random 32-bit number
	{
		register int j;
		register unsigned int new_rand;

		// the essential algorithm: XOR together the value at the pointer
		// and the value 147 before. Store that value at the pointer, and 
		// return it.
		
		// select the other value at 147 earlier (or 103 later; same thing)
		if (mnIndex >= 147) {
			j = mnIndex - 147;	/* Wrap pointer around */
		} else {
			j = mnIndex + 103;
		}

		// XOR the values
		new_rand = maunTable[mnIndex] ^ maunTable[j];

		// stash it at the pointer
		maunTable[mnIndex] = new_rand;

		// increment pointer for next time
		if (mnIndex >= 249) {		
			mnIndex = 0;
		} else {
			mnIndex++;
		}

		return new_rand;
	}
};

//
// static r250
//

void r250_init(unsigned int unSeed);
unsigned int r250();
double dr250();

//
// static linear congruent
//

void vLCSeed(unsigned int unSeed);
unsigned int unLCRand(void);
double dLCRand(void);

#endif